from collections import OrderedDict

from cassandra.cqlengine import columns
from cassandra.cqlengine.models import (
    Model as CqlEngineModel,
    ModelMetaClass as CqlEngineModelMetaClass
)

from djangotoolbox.fields import (
    ListField,
    SetField,
    DictField
)

from .exception import DatabaseError


internal_type_to_column_map = {
    'AutoField': columns.Integer,
    'AutoFieldUUID': columns.UUID,
    'RelatedAutoField': columns.UUID,
    'ForeignKey': columns.UUID,
    'OneToOneField': columns.UUID,
    'ManyToManyField': columns.UUID,
    'BigIntegerField': columns.BigInt,
    'BooleanField': columns.Boolean,
    'CharField': columns.Text,
    'CommaSeparatedIntegerField': columns.Text,
    'DateField': columns.Date,
    'DateTimeField': columns.DateTime,
    'DecimalField': columns.Decimal,
    'EmailField': columns.Text,
    'FileField': columns.Text,
    'FilePathField': columns.Text,
    'FloatField': columns.Float,
    'ImageField': columns.Text,
    'IntegerField': columns.Integer,
    'IPAddressField': columns.Inet,
    'NullBooleanField': columns.Boolean,
    'PositiveIntegerField': columns.Integer,
    'PositiveSmallIntegerField': columns.Integer,
    'SlugField': columns.Text,
    'SmallIntegerField': columns.Integer,
    'TextField': columns.Text,
    'TimeField': columns.DateTime,
    'URLField': columns.Text,
    'AbstractIterableField': columns.List,
    'ListField': columns.List,
    'SetField': columns.Set,
    'DictField': columns.Map,
    'EmbeddedModelField': columns.Map,
    'RawField': columns.Blob,
    'BlobField': columns.Blob
}


def get_cql_column_type(field):
    internal_type = field.get_internal_type()
    if internal_type == 'ForeignKey':
        internal_type = (
            field.related.model._meta.pk.get_internal_type()
        )

    return internal_type_to_column_map[
        internal_type
    ]


class CqlColumnFamilyMetaClass(CqlEngineModelMetaClass):
    __column_families__ = {}

    @staticmethod
    def update_column_family(column_family):
        assert issubclass(column_family, CqlEngineModel)
        CqlColumnFamilyMetaClass.__column_families__[
            column_family.__name__
        ] = column_family

    def __new__(
        meta,
        name,
        bases,
        attrs
    ):
        ordered_attrs = OrderedDict(attrs)

        if name in CqlColumnFamilyMetaClass.__column_families__:
            return CqlColumnFamilyMetaClass.__column_families__[name]

        model = attrs.get('__model__')
        registered_model = model
        try:
            from django.apps import apps
            registered_model = apps.get_model(
                model._meta.app_label,
                model._meta.model_name
            )

        except:
            pass

        if hasattr(registered_model, 'Cassandra'):
            cassandra_options = registered_model.Cassandra

        else:
            cassandra_options = None

        if None is not model:
            '''
            Create partition/clustering keys first.
            '''
            partition_keys = []
            if (
                cassandra_options and
                hasattr(cassandra_options, 'partition_keys')
            ):
                partition_keys = cassandra_options.partition_keys

            primary_field_name = None

            if partition_keys:
                for column_name in partition_keys:
                    field = model._meta.get_field(column_name)
                    field_name = (
                        field.db_column if
                        field.db_column else
                        field.column
                    )

                    column_type = get_cql_column_type(field)
                    column = column_type(
                        primary_key=True,
                        partition_key=True,
                        db_field=field_name
                    )
                    ordered_attrs[
                        field.db_column if
                        field.db_column else
                        field.column
                    ] = column

            else:
                primary_key_field = model._meta.pk
                primary_field_name = (
                    primary_key_field.db_column
                    if primary_key_field.db_column
                    else primary_key_field.column
                )

                column_type = get_cql_column_type(primary_key_field)

                column = column_type(
                    primary_key=True,
                    partition_key=True,
                    db_field=primary_field_name
                )
                ordered_attrs[
                    primary_key_field.db_column if
                    primary_key_field.db_column else
                    primary_key_field.column
                ] = column

            clustering_keys = []
            if (
                cassandra_options and
                hasattr(cassandra_options, 'clustering_keys')
            ):
                clustering_keys = cassandra_options.clustering_keys

                for column_name in clustering_keys:
                    field = model._meta.get_field(column_name)
                    field_name = (
                        field.db_column if
                        field.db_column else
                        field.column
                    )

                    if (
                        field_name in partition_keys or
                        field_name == primary_field_name
                    ):
                        # Skip primary key or partition keys if included here.
                        continue

                    column_type = get_cql_column_type(field)
                    column = column_type(
                        primary_key=True,
                        db_field=field_name
                    )
                    ordered_attrs[
                        field.db_column if
                        field.db_column else
                        field.column
                    ] = column

            for field in model._meta.fields:
                if 'Token' == field.get_internal_type():
                    continue

                field_name = (
                    field.db_column if
                    field.db_column else
                    field.column
                )
                if (
                    field.primary_key or
                    field_name in partition_keys or
                    field_name in clustering_keys
                ):
                    continue

                column_type = get_cql_column_type(field)

                args = []
                if (
                    isinstance(field, DictField)
                ):
                    args.append(columns.Text)
                    args.append(
                        get_cql_column_type(field.item_field)
                    )

                elif (
                    isinstance(field, ListField) or
                    isinstance(field, SetField)
                ):
                    args.append(
                        get_cql_column_type(field.item_field)
                    )

                column = column_type(
                    index=field.db_index,
                    db_field=field_name,
                    required=not field.blank,
                    *args
                )

                ordered_attrs[field_name] = column

        column_family = super(CqlColumnFamilyMetaClass, meta).__new__(
            meta,
            name,
            bases,
            ordered_attrs
        )

        CqlColumnFamilyMetaClass.__column_families__[name] = column_family
        return column_family


class CqlColumnFamily(CqlEngineModel):
    __model__ = None
    __abstract__ = True
    __metaclass__ = CqlColumnFamilyMetaClass


'''
Table Options:

'comment': None,
'read_repair_chance': None,
'dclocal_read_repair_chance': None,
'replicate_on_write': None,
'gc_grace_seconds': None,
'bloom_filter_fp_chance': None,
'caching': None,
'compaction_strategy_class': None,
'compaction_strategy_options': None,
'min_compaction_threshold': None,
'max_compaction_threshold': None,
'compression_parameters': None,
'min_index_interval': None,
'max_index_interval': None,
'index_interval': None,
'speculative_retry': None,
'rows_per_partition_to_cache': None,
'memtable_flush_period_in_ms': None,
'populate_io_cache_on_flush': None,
'compaction': None,
'compression': None,
'default_time_to_live': None
'''

default_table_options = {
}


default_cassandra_model_settings = {
    'table_options': default_table_options
}


def get_column_family(
    connection,
    model
):
    connection_settings = connection.settings_dict

    registered_model = model
    try:
        from django.apps import apps
        registered_model = apps.get_model(
            model._meta.app_label,
            model._meta.model_name
        )

    except:
        pass

    if hasattr(registered_model, 'Cassandra'):
        cassandra_options = registered_model.Cassandra

    else:
        cassandra_options = default_cassandra_model_settings

    if hasattr(cassandra_options, 'keyspace'):
        keyspace = cassandra_options.keyspace

    else:
        keyspace = connection_settings.get('DEFAULT_KEYSPACE')

    keyspace_settings = connection_settings.get('KEYSPACES', {}).get(keyspace)
    if None is keyspace_settings:
        keyspace_settings = {}  # Replace with default keyspace settings.

    table_options = default_table_options

    if hasattr(cassandra_options, 'table_options'):
        if not isinstance(cassandra_options.table_options, dict):
            raise DatabaseError(
                'The value of table_options in the Cassandra class '
                'must be a dict containing overrides for the default'
                'column family options.'
            )
        table_options.update(cassandra_options.table_options)

    return type(
        str(model._meta.db_table),
        (CqlColumnFamily,), {
            '__model__': model,
            'table_options': table_options
        }
    )



#   Copyright 2010 BSN, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import re
import itertools
import warnings

from collections import OrderedDict

from .exceptions import InefficientQueryError

from .utils import sort_rows

SECONDARY_INDEX_SUPPORT_ENABLED = True


class InvalidSortSpecException(Exception):
    def __init__(self):
        super(InvalidSortSpecException, self).__init__(
            'The row sort spec must be a sort spec '
            'tuple/list or a tuple/list of sort specs'
        )


class InvalidRowCombinationOpException(Exception):
    def __init__(self):
        super(InvalidRowCombinationOpException, self).__init__(
            'Invalid row combination operation'
        )


class InvalidPredicateOpException(Exception):
    def __init__(self):
        super(InvalidPredicateOpException, self).__init__(
            'Invalid/unsupported query predicate operation'
        )


COMPOUND_OP_AND = 1
COMPOUND_OP_OR = 2


class RangePredicate(object):
    def __init__(
        self,
        column,
        start=None,
        start_inclusive=True,
        end=None,
        end_inclusive=True
    ):
        self.column = column
        self.start = start
        self.start_inclusive = start_inclusive
        self.end = end
        self.end_inclusive = end_inclusive

    def __repr__(self):
        s = '(RANGE: '
        if self.start:
            op = '<=' if self.start_inclusive else '<'
            s += (unicode(self.start) + op)
        s += self.column
        if self.end:
            op = '>=' if self.end_inclusive else '>'
            s += (op + unicode(self.end))
        s += ')'
        return s

    def _is_exact(self):
        return (
            (self.start is not None) and
            (self.start == self.end) and
            self.start_inclusive and
            self.end_inclusive
        )

    def can_evaluate_efficiently(
        self,
        partition_columns,
        clustering_columns,
        indexed_columns
    ):
        if self._is_exact():
            return self.column in itertools.chain(
                ['pk__token'],
                partition_columns,
                clustering_columns,
                indexed_columns
            )

        else:
            return self.column in itertools.chain(
                ['pk__token'],
                clustering_columns
            )

    def incorporate_range_op(self, column, op, value, parent_compound_op):
        if column != self.column:
            return False

        # FIXME: The following logic could probably be tightened up a bit
        # (although perhaps at the expense of clarity?)
        if parent_compound_op == COMPOUND_OP_AND:
            if op == 'gt':
                if self.start is None or value >= self.start:
                    self.start = value
                    self.start_inclusive = False
                    return True
            elif op == 'gte':
                if self.start is None or value > self.start:
                    self.start = value
                    self.start_inclusive = True
                    return True
            elif op == 'lt':
                if self.end is None or value <= self.end:
                    self.end = value
                    self.end_inclusive = False
                    return True
            elif op == 'lte':
                if self.end is None or value < self.end:
                    self.end = value
                    self.end_inclusive = True
                    return True
            elif op == 'exact' or op == 'eq':
                if self._matches_value(value):
                    self.start = self.end = value
                    self.start_inclusive = self.end_inclusive = True
                    return True
            elif op == 'startswith':
                # For the end value we increment the ordinal value of the
                # last character in the start value and make the end value
                # not inclusive
                end_value = value[:-1] + chr(ord(value[-1])+1)
                if (
                    ((self.start is None) or (value > self.start)) and
                    ((self.end is None) or (end_value <= self.end))
                ):
                    self.start = value
                    self.end = end_value
                    self.start_inclusive = True
                    self.end_inclusive = False
                    return True
            else:
                raise InvalidPredicateOpException()
        elif parent_compound_op == COMPOUND_OP_OR:
            if op == 'gt':
                if self.start is None or value < self.start:
                    self.start = value
                    self.start_inclusive = False
                    return True
            elif op == 'gte':
                if self.start is None or value <= self.start:
                    self.start = value
                    self.start_inclusive = True
                    return True
            elif op == 'lt':
                if self.end is None or value > self.end:
                    self.end = value
                    self.end_inclusive = False
                    return True
            elif op == 'lte':
                if self.end is None or value >= self.end:
                    self.end = value
                    self.end_inclusive = True
                    return True
            elif op == 'exact' or op == 'eq':
                if self._matches_value(value):
                    return True
            elif op == 'startswith':
                # For the end value we increment the ordinal value of the
                # last character in the start value and make the end value
                # not inclusive
                end_value = value[:-1] + chr(ord(value[-1])+1)
                if (
                    ((self.start is None) or (value <= self.start)) and
                    ((self.end is None) or (end_value > self.end))
                ):
                    self.start = value
                    self.end = end_value
                    self.start_inclusive = True
                    self.end_inclusive = False
                    return True
        else:
            raise InvalidPredicateOpException()

        return False

    def _matches_value(self, value):
        if value is None:
            return False
        if self.start is not None:
            if self.start_inclusive:
                if value < self.start:
                    return False
            elif value <= self.start:
                return False

        if self.end is not None:
            if self.end_inclusive:
                if value > self.end:
                    return False
            elif value >= self.end:
                return False
        return True

    def row_matches(self, row):
        value = row.get(self.column, None)
        return self._matches_value(value)

    def get_matching_rows(self, query):
        rows = query.get_row_range(self)
        return rows


class OperationPredicate(object):
    def __init__(self, column, op, value=None):
        self.column = column
        self.op = op
        self.value = value
        if op == 'regex' or op == 'iregex':
            flags = re.I if op == 'iregex' else 0
            self.pattern = re.compile(value, flags)

    def __repr__(self):
        return '(OP: ' + self.op + ':' + unicode(self.value) + ')'

    def can_evaluate_efficiently(
        self,
        pk_column,
        clustering_columns,
        indexed_columns
    ):
        return False

    def row_matches(self, row):
        row_value = row.get(self.column, None)
        if self.op == 'isnull':
            return row_value is None
        # FIXME: Not sure if the following test is correct in all cases
        if (row_value is None) or (self.value is None):
            return False
        if self.op == 'in':
            return row_value in self.value
        if self.op == 'istartswith':
            return row_value.lower().startswith(self.value.lower())
        elif self.op == 'endswith':
            return row_value.endswith(self.value)
        elif self.op == 'iendswith':
            return row_value.lower().endswith(self.value.lower())
        elif self.op == 'iexact':
            return row_value.lower() == self.value.lower()
        elif self.op == 'contains':
            return row_value.find(self.value) >= 0
        elif self.op == 'icontains':
            return row_value.lower().find(self.value.lower()) >= 0
        elif self.op == 'regex' or self.op == 'iregex':
            return self.pattern.match(row_value) is not None
        else:
            raise InvalidPredicateOpException()

    def incorporate_range_op(self, column, op, value, parent_compound_op):
        return False

    def get_matching_rows(self, query):
        # get_matching_rows should only be called for predicates that can
        # be evaluated efficiently, which is not the case for
        # OperationPredicate's
        raise NotImplementedError(
            'get_matching_rows() called for inefficient predicate'
        )


class CompoundPredicate(object):
    def __init__(self, op, negated=False, children=None):
        self.op = op
        self.negated = negated
        self.children = children
        if self.children is None:
            self.children = []

    def __repr__(self):
        s = '('
        if self.negated:
            s += 'NOT '
        s += ('AND' if self.op == COMPOUND_OP_AND else 'OR')
        s += ': '
        first_time = True
        if self.children:
            for child_predicate in self.children:
                if first_time:
                    first_time = False
                else:
                    s += ','
                s += unicode(child_predicate)
        s += ')'
        return s

    def can_evaluate_efficiently(
        self,
        pk_column,
        clustering_columns,
        indexed_columns
    ):
        if self.negated:
            return False

        if self.op == COMPOUND_OP_AND:
            for child in self.children:
                if not child.can_evaluate_efficiently(
                    pk_column,
                    clustering_columns,
                    indexed_columns
                ):
                    return False

            return True

        elif self.op == COMPOUND_OP_OR:
            for child in self.children:
                if not child.can_evaluate_efficiently(
                        pk_column,
                        clustering_columns,
                        indexed_columns
                ):
                    return False

            return True

        else:
            raise InvalidPredicateOpException()

    def row_matches_subset(self, row, subset):
        if self.op == COMPOUND_OP_AND:
            matches = True
            for predicate in subset:
                if not predicate.row_matches(row):
                    matches = False
                    break

        elif self.op == COMPOUND_OP_OR:
            matches = False
            for predicate in subset:
                if predicate.row_matches(row):
                    matches = True
                    break

        else:
            raise InvalidPredicateOpException()

        if self.negated:
            matches = not matches

        return matches

    def row_matches(self, row):
        return self.row_matches_subset(row, self.children)

    def incorporate_range_op(self, column, op, value, parent_predicate):
        return False

    def add_filter(self, column, op, value):
        column_name = (
            column.db_column
            if column.db_column
            else column.column
        )
        if op in ('lt', 'lte', 'gt', 'gte', 'eq', 'exact', 'startswith'):
            if not len(self.children):
                child = RangePredicate(column_name)
                incorporated = child.incorporate_range_op(
                    column_name,
                    op,
                    value,
                    COMPOUND_OP_AND
                )
                assert incorporated
                self.children.append(child)

            else:
                incorporated = None
                for child in self.children:
                    incorporated = child.incorporate_range_op(
                        column_name,
                        op,
                        value,
                        self.op
                    )
                    if incorporated:
                        break

                if not incorporated:
                    child = RangePredicate(column_name)
                    incorporated = child.incorporate_range_op(
                        column_name,
                        op,
                        value,
                        COMPOUND_OP_AND
                    )
                    assert incorporated
                    self.children.append(child)
        else:
            child = OperationPredicate(column_name, op, value)
            self.children.append(child)

    def add_child(self, child_query_node):
        self.children.append(child_query_node)

    def get_matching_rows(self, query):
        pk_column = query.query.get_meta().pk.column

        # In the first pass we handle the query nodes that can be processed
        # efficiently. Hopefully, in most cases, this will result in a
        # subset of the rows that is much smaller than the overall number
        # of rows so we only have to run the inefficient query predicates
        # over this smaller number of rows.
        range_predicates = []
        inefficient_predicates = []
        result = None
        for predicate in self.children:
            if predicate.can_evaluate_efficiently(
                pk_column,
                query.clustering_columns,
                query.filterable_columns
            ):
                range_predicates.append(predicate)

            else:
                inefficient_predicates.append(predicate)

        cql_query = query.get_row_range(range_predicates)

        if query.ordering:
            for order in query.ordering:
                cql_query = cql_query.order_by(order)

        result = cql_query

        if None is result:
            result = []

        else:
            result = [OrderedDict(zip(
                query.column_names,
                row
            )) for row in result]

        if (
            inefficient_predicates or
            query.inefficient_ordering
        ):
            if not query.allows_inefficient:
                raise InefficientQueryError(query)

            warnings.warn(InefficientQueryError.message)

        if inefficient_predicates:
            result = [
                row for row in result if self.row_matches_subset(
                    row,
                    inefficient_predicates
                )
            ]

        if query.inefficient_ordering:
            for order in query.inefficient_ordering:
                sort_rows(result, order)

        if query.low_mark is not None or query.high_mark is not None:
            result = result[query.low_mark:query.high_mark]

        return result



from django.db.models import (
    Model as DjangoModel,
    Manager
)

from .fields import TokenPartitionKeyField
from .query import QuerySet


class ColumnFamilyManager(Manager.from_queryset(QuerySet)):
    pass


class ColumnFamilyModel(DjangoModel):
    class Meta:
        abstract = True

    pk_token = TokenPartitionKeyField()

    def __init__(
        self,
        *args,
        **kwargs
    ):
        super(ColumnFamilyModel, self).__init__(
            *args,
            **kwargs
        )

    objects = ColumnFamilyManager()
